/*
Example of some Clipper 5.x code to read messages from a QWK-file.
It's not complete, but should give you a pretty good handle on how to
extract the messages from a QWK-file and place them in a DBF.
This code is close to functional, with the exception of the OOP-stuff that
isn't included. Rewriting this to standard Clipper code shouldn't
be all that hard though.

Do with it what you want. With, or without thanking me :)

A QWK-file is nothing more or less then an archive containing a number of
specific files. Most QWK-files are created with PKZIP, that's why this
code presumes the QWK can be unarchived with PKUNZIP.

The only two files contained in the QWK-archive, that are of interest
to this piece of code are CONTROL.DAT and MESSAGES.DAT.
The CONTROL.DAT contains information about the BBS you got the QWK-file from
and information about the conferences.
The MESSAGES.DAT contains the actual messages.

The layout of these two files is specified in QWKLAY16.ZIP archive that I'll
include with this file, but also can be derived from looking at this code.

Not all information contained in the two DAT-files is used.
All the other files in the QWK-archive are discarded since they are
not essential to move messages from a QWK to a DBF.


This file contains the following functions;

Name        Parameter   What does it do
-----------------------------------------------------------------------------
Main        cFile       Nothing really :). Calls QWKRead(cFile)
QWKRead     cFile       Does some checks and
                           Unarchives the QWK-file in a temp work dir
                           Calls ReadContr() to read the CONTROL.DAT
                           Calls ReadMess()  to read the MESSAGES.DAT
ReadContr   [nHCT]      Reads the CONTROL.DAT file. Stores the relevant
                        information in STATIC arrays.
                        With the right parameter (see #define's)
                        can be used to retrieve that information.
ReadMess                Reads the MESSAGES.DAT file from beginning till end
                        and stores the information read into a messages.dbf
                        file.




The following functions are used, but not contained in this source-file;

Name        What does it do
-----------------------------------------------------------------------------
GETDDOBJ    Create an object for accessing a DBF file.
            Loop at Std. Clipper functions DBUseArea() and DBAppend().
BINFILE     Create an object for accessing a file at 'binary' (low) level
            Loop at Std. Clipper functions FOpen(), FRead() and FClose().
            (The ReadLine() method does an FRead() up to and including the next
             CHR(13)+CHR(10) eol-pointer and returns that line exclusive the
             eol-pointer)
DOS         Wrapper function for a call to Blinker's SWPRUNCMD().
            Executes a DOS-command (PKUNZIP in this case)
            Although I would advise strongly against it's use, the Std Clipper
            RUN command is somewhat equivilant.
NICE_DIR    Function that add's a trailing backslash and inserts full path
            (including drive letter etc) if not present.
INFO        Your basic warning-window function.

MKDIR       Funcky II-function to create a directory.
CHDIR       Funcky II-function to change the current directory.
RMDIR       Funcky II-function to remove a directory.
FERASEALL   Funcky II-function to erase all the files that match the pattern.

IF you need any more help, you might be able to contact me at;

anne@hacom.nl

Anne Krijger

*/

#include "inkey.ch"
#include "common.ch"
#include "set.ch"

#define CONTROL_HEADER    1
#define CONTROL_CONFS     2
#define CONTROL_TRAILER   3


FUNCTION Main(cFile)
// cFile is the QWK-file name as command-prompt parameter //

QWKRead(cFile)

/*
Do whatever you want with the messages database here.
*/

RETURN TRUE


FUNCTION QWKRead(cFile)
/*

AJR\ is the temp work dir, needed for unarchiving the QWK-archive.
Can be replaced by a variable, and any other name,
as can the other references to specific directories.

*/
LOCAL cQWKPath, cCurDir := CurDir()
cQWKPath   := "C:\COMMS\KQWK\QWK\"
IF !FILE("AJR\*.DAT")                                      // When the temp work dir does not contain any DAT-files //
   IF FILE(cFile)                                          // Check if the file exists //
      mkdir("AJR")                                         // Create temp-dir //
      chdir("AJR")                                         //Make it the active dir //
      IF Dos("Pkunzip "+ALLTRIM(cFile)+" -3 -O", 0, Nice_Dir(cCurDir)+"AJR")==0
         // If DOS-command PKZIP was executed succesfully //

         chdir("..")                                       // Return to one lower level in dir structure //

         ReadContr()                                       // Read the control.dat file //
         ReadMess()                                        // Read the messages in the messages.dat file //

      ELSE                                                 // If something went wrong, don't forget to... //
         chdir("..")                                       // Return to one lower level in dir structure //
      ENDIF

      IF FILE("AJR\*.DAT")                                 // This should always be true //
         IF !(FEraseAll("AJR\*.*")>0)                      // If no files are deleted by this Funcky-function //
            Info("No File deleted met feraseall('AJR\*.*') !", "W05ET") // Give warning //
         ENDIF
      ENDIF
      rmdir("AJR")                                         // And remove the temp work dir again //

   ELSE                                                    // If not, give warning //
      Info(cFile+" not found in the "+cQWKPath+" dir!", "W05ET")
   ENDIF
ELSE                                                       // If the temp work dir is not empty //
   // Give a warning //
   Info("Work directory still contains files.|Unable to unarchive the QWK-file.", "WET")
ENDIF
RETURN TRUE





FUNCTION ReadContr(nHCT)
/*
Either read control.dat file,
or return the requested parts of last read.

The information gathered from the control.dat is stored in three different STATIC array's.

*/
STATIC acHeader[11], acConfs := {}, acTrailer := {}

LOCAL nLen, oFile, _SetDate, i := 0
LOCAL xRet := NIL, cLine := ""
IF PCOUNT()==0                                             // No parameters; read the Control.dat file //
   IF FILE("AJR\CONTROL.DAT")                              // If the file exists //

      oFile := BinFile():New(0, "AJR\CONTROL.DAT")         // OOP-handle to the file, read-only, shareable //
      IF oFile:Open()==0                                   // If opened correctly //

         _SetDate := SET(_SET_DATEFORMAT, 'MM-DD-YY')      // Dates in QWK-files are american format !! //
         oFile:GoTop()                                     // Go to BOF //

         // Read the header-lines 11 in total //
         FOR i := 1 TO 11                                  // 11 is the defined number of header lines. See QWK-spec //
            acHeader[i] := oFile:ReadLine()                // Read Header, 11 lines long //
         NEXT

         // Read the conferences. 2 lines per conference. //
         // The (total number of conferences)-1 is listed on the 11th line of the header //
         // Add one to the number of conferences; Main Board is never counted. See QWK-spec //
         nLen    := VAL(acHeader[11])+1
         acConfs := ARRAY(nLen)                            // Create conferences array (will be a problem if over 4096 conferences !!!) //
         FOR i := 1 TO nLen
            /*
            Conferences-array is filled with the following structure;
               Conference-number C 4
               Delimiter ('|')
               Conference-name   C (unlimited, but 15 by QWKs-spec)

            Read two lines of the control.dat at a time.
            First is the conference number, second line contians the conference name.
            */
            acConfs[i] := PADL(ALLTRIM(oFile:ReadLine()),4)+"|"+ALLTRIM(oFile:ReadLine())
         NEXT

         i := 0                                            // Set counter to 0 //
         DO WHILE !oFile:IsEof()                           // read the rest of the control.dat file //
            cLine := ALLTRIM(oFile:ReadLine())             // One line at a time //
            IF !(cLine=="0")                               // Control.dat can be filled out with lines containing just "0" //
               AADD(acTrailer, cLine)                      // If not, add the contents of the line to the Trailer-array (not used here) //
               i++                                         // Keep count of the number of 'extra lines' read //
            ENDIF
         ENDDO
         oFile:Close()                                     // Close the control.dat file //
         SET(_SET_DATEFORMAT, _SetDate)                    // and set date format back to the way we are used to :) //

      ELSE                                                    // If not opened correctly //
         Info("Error, no control.dat. Thank you...", "W03ET") // Give warning //
      ENDIF

   ELSE                                                    // If not found //
      Info("Error, no control.dat. Thank you...", "W03ET") // Give warning //
   ENDIF

ELSEIF ISNUMBER(nHCT)                                        // Header-array, Conference-array of Trailer-array requested //
   DO CASE
      CASE nHCT==CONTROL_HEADER                            // Header-array requested //
         xRet := acHeader

      CASE nHCT==CONTROL_CONFS                             // Conference-array requested //
         xRet := acConfs

      CASE nHCT==CONTROL_TRAILER                           // Trailer-array (extra files) requested //
         xRet := acTrailer

   ENDCASE
ENDIF
RETURN xRet





FUNCTION ReadMess
/*

This function actually reads the messages from the MESSAGES.DAT file.
It's done in a 'brute force' way :) From the beginning of the file till
the end, without caring anything about the index-files that also might
be contained in the QWK-archive.

*/
LOCAL nBytes := 128, nBlocks := 0, cText := "", cLine
LOCAL cStatus := "", cMessage := "", dDate, cTime := ""
LOCAL cTo := "", cFrom := "", cSubject := "", cMessRef := ""
LOCAL lActive := FALSE, cConf := ""
LOCAL nTotMes := 0, nPos := 0, i := 0, nKey
LOCAL oFile, _SetDate
LOCAL acHeader := {}, acConfs := {}, acTrailer := {}
LOCAL oMess

oMess     := GetDDOBJ("messages")                          // OOP-handle to the messages.dbf //

acHeader  := ReadContr(CONTROL_HEADER)                     // Get the header info from the Control.dat that was read //
acConfs   := ReadContr(CONTROL_CONFS)                      // Get the conferences info from the Control.dat that was read //
acTrailer := ReadContr(CONTROL_TRAILER)                    // Get the extra (trailer) info from the Control.dat that was read //

IF FILE("AJR\MESSAGES.DAT")                                // If the MESSAGES.DAT file exists in the temp work dir //
   oFile := BinFile():new(0, "AJR\MESSAGES.DAT")           // OOP-handle to the file, read-only, shareable //
   cLine := SPACE(nBytes)                                  // Standard lenght 1 block (QWK-Spec) //

   IF oFile:Open()==0                                      // If opened correctly //
      oFile:GoTop()                                        // Go to BOF //
      oFile:Read(@cLine, nBytes)                           // Read first block of nBytes in messages.dat. (Is of no relevance (copyright notice)) //

      _SetDate := SET(_SET_DATEFORMAT, 'MM-DD-YY')         // Dates in QWK-files are american format !! //

      nKey := 0

      DO WHILE !(nKey==K_ESC)                              // Make  it possible to 'escape'
         cLine := SPACE(nBytes)                            // Standard lenght 1 block //

         IF oFile:Read(@cLine, nBytes)                     // First block of nBytes per message contains it's Header //
            cStatus  := SUBSTR(cLine, 1, 1)
            cMessage := SUBSTR(cLine, 2, 7)
            dDate    := CTOD(SUBSTR(cLine, 9, 8))
            cTime    := SUBSTR(cLine, 17, 5)
            cTo      := SUBSTR(cLine, 22, 25)
            cFrom    := SUBSTR(cLine, 47, 25)
            cSubject := SUBSTR(cLine, 72, 25)
            cMessRef := PADL(ALLTRIM(SUBSTR(cLine, 109, 8)),8)
            nBlocks  := VAL(SUBSTR(cLine, 117, 6))         // Lenght in number of blocks of nBytes of this message. INCULDING the just read header ! //
            lActive  := SUBSTR(cLine, 123, 1)==CHR(225)
            cConf    := PADL(ALLTRIM(STR(BIN2W(SUBSTR(cLine, 124, 2)))), 4)
            nTotMes++                                      // Count total number of messages read //

            cText     := ""

            nBlocks--                                      // First Block is already read. Therefor one less remains to be read for this message //
            cLine := SPACE(nBlocks*nBytes)                 // Set up reference var to read the entire messagetext at once in //

            IF oFile:Read(@cLine, (nBlocks*nBytes))        // Read nBlocks of nBytes; Eg: the message-text //

               cText += RTRIM(cLine)                       // Remove trailing spaces //
               @ 24, 0 SAY "Reading Message : "+STR(nTotMes) // Show that something is happening //

            ENDIF

            IF !EMPTY(cText)                               // When something actually is read //
               IF oMess:Append()                           // Append a record to the message-base //
                  // And replace the right fields with the data of this message //

                  REPLACE messages->status     WITH cStatus

                  // text is a memo-field //
                  REPLACE messages->text       WITH cText

                  REPLACE messages->message    WITH cMessage
                  REPLACE messages->ref        WITH cMessRef

                  REPLACE messages->date       WITH dDate
                  REPLACE messages->time       WITH cTime

                  REPLACE messages->to         WITH cTo
                  REPLACE messages->from       WITH cFrom
                  REPLACE messages->subject    WITH cSubject

                  IF !EMPTY(cConf)
                     nEl := ASCAN(acConfs, cConf)          // If conferencenumber found in conference-array //
                     IF nEl > 0
                        REPLACE messages->confname WITH TOKEN(acConfs[nEl], "|", 2) // get the conferencename and store it //
                     ENDIF
                  ENDIF

               ENDIF
            ENDIF

         ENDIF

         IF oFile:IsEof()                                  // When at the end of the messages.dat file //
            EXIT                                           // Exit the DO-WHILE-loop //
         ENDIF

         nKey := LASTKEY()                                 // See if someone is hitting <ESC> like a madman :) //

      ENDDO
      oFile:Close()                                        // Close the messages.dat file //
      SET(_SET_DATEFORMAT, _SetDate)                       // and set date format back to the way we are used to :) //

   ELSE                                                    // If not opened correctly //
      Info("Error, no messages.dat. Thank you...", "W03ET")// Give warning //
   ENDIF
ELSE                                                       // If not opened correctly //
   Info("Error, no messages.dat. Thank you...", "W03ET")   // Give warning //
ENDIF

RETURN TRUE
